// cWorld.cpp

#include "cWorld.h"

#include "cDome.h"
#include "cObject.h"

#define MMTIMER

#if defined(_WIN32) && defined(MMTIMER)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#else
#include <sys/timeb.h>
#endif


float GetSPF(unsigned long *newClock, unsigned long *oldClock)
{
  float spf;

#if defined(_WIN32) && defined(MMTIMER)
  // win only
  int ok;
  ok = 0;
  if (timeBeginPeriod(1) == TIMERR_NOERROR) ok = 1;
  newClock[0] = timeGetTime();
  if (ok) timeEndPeriod(1);
  spf = (float) (newClock[0] - oldClock[0]) / 1000.0f;
#else
  // Possibly this will work on every system?
  struct timeb timb;
  ftime(&timb);
  newClock[0] = timb.millitm + 1000 * timb.time;
  spf = (float) ((double) (newClock[0] - oldClock[0]) / 1000.0);
  /*
  // Works everywhere - but only gets process time (sux on win).
  newClock[0] = clock();
  spf = (float) (newClock[0] - oldClock[0]) / CLOCKS_PER_SEC;
  */
#endif
  oldClock[0] = newClock[0];
  return spf;
}


cWorld* cWorld::sWorld = NULL;

cWorld::cWorld()
{
  mGravity[X] = 0;
  mGravity[Y] = 9.8f;
  mGravity[Z] = 0;

  mBufferedCorpses = &mCorpses[0];

  updateTimers();
  updateTimers(); // yes, two times!!
  
  mFPS = 30.0f;
  mSPF = 1.0f / mFPS;
  mTime = 0.0f;
  
  cWorld::sWorld = this;
  
  //mod = new cDummyModel;
  //mod->setName("cDummyModel");
  //planet->spawnObject(mod);

  mParticleModel = new cParticleModel;
  //mParticleModel->setName("cParticleModel");
  //this->spawnObject(mParticleModel);

  mConstructModel = new cConstructModel;
  //mConstructModel->setName("cConstructModel");
  //this->spawnObject(mConstructModel);

  mPlantModel = new cPlantModel;
  //mPlantModel->setName("cPlantModel");
  //this->spawnObject(mPlantModel);
}


// Spawning And Fragging


void cWorld::spawnObject(cObject *object)
{
  object->mWorld = this;
  mObjects.push_back(object);
  object->onSpawn();
}


void cWorld::fragObject(cObject *object)
{
  if (object->mWorld != this) return;
  if (object->mFragged) return;
  object->mFragged = true;
  mBufferedCorpses->push_back(object);
  object->onFrag();
}


void cWorld::bagFragged()
{
  // Switch corpse-double-buffer.
  mBufferedCorpses = (mBufferedCorpses == &mCorpses[0]) ? &mCorpses[1] : &mCorpses[0];
  
  // Finally delete objects which where scheduled to be
  // deleted in the last frame.
  while (!mBufferedCorpses->empty())
  {
    cObject* object = mBufferedCorpses->front();
    mBufferedCorpses->pop_front();
    mObjects.remove(object);
    delete object;
  };
}


// Simulation Step


void cWorld::updateTimers()
{
  mSPF = GetSPF(&mNewClock, &mOldClock);
  mFPS = 1.0 / (mSPF + 0.000001f);
  mTime += mSPF;
}


void cWorld::animateSelf(float spf)
{
  loopObjects(mObjects)
  {
    cObject* object = *it;
    it++;
    if (object->mFragged) continue;
    object->animateSelf(spf);
  };
}


void cWorld::transformSelf()
{
  loopObjects(mObjects)
  {
    cObject* object = *it;
    it++;
    if (object->mFragged) continue;
    glPushMatrix();
      object->transformSelf();
    glPopMatrix();
  };
}


void cWorld::drawSelf()
{
  loopObjects(mObjects)
  {
    cObject* object = *it;
    it++;
    if (object->mFragged) continue;
    glPushMatrix();
      object->drawSelf();
    glPopMatrix();
  };
}


void cWorld::drawEffect()
{
  loopObjects(mObjects)
  {
    cObject* object = *it;
    it++;
    if (object->mFragged) continue;
    glPushMatrix();
      object->drawEffect();
    glPopMatrix();
  };
}


// World Scanning And Filtering


void cWorld::printNames(list<cObject*>* objects)
{
  char buffer[1024];
  string s = "[";
  if (objects == NULL) objects = &mObjects;
  loopObjects(*objects)
  {
    sprintf(buffer, "%lu", (*it)->mRoles);
    s += " ";
    s += *( (*it)->mName );
    s += "(";
    s += buffer;
    s += ")";
  }
  s += " ]";
  printf("%s\n", s.c_str());
}


list<cObject*>* cWorld::filterByRole(cObject* ex, unsigned long rolemask, bool all, int maxamount, list<cObject*>* objects)
{
  list<cObject*>* result = new list<cObject*>;
  int amount = maxamount;
  if (objects == NULL) objects = &mObjects;
  loopObjects(*objects)
  {
    if (amount == 0) break;
    cObject *object = *it; it++;
    unsigned long intersection = object->mRoles & rolemask;
    if ((intersection != rolemask) && !((intersection != 0) && !all)) continue;
    if (object->mFragged) continue;
    if (object == ex) continue;
    amount--;
    result->push_back(object);
  }
  return result;
}


list<cObject*>* cWorld::filterByRange(cObject* ex, float* origin, float minrange, float maxrange, int maxamount, list<cObject*>* objects)
{
  list<cObject*>* result = new list<cObject*>;
  int amount = maxamount;
  if (objects == NULL) objects = &mObjects;
  loopObjects(*objects)
  {
    if (amount == 0) break;
    cObject *object = *it; it++;
    if (object->mFragged) continue;
    if (object == ex) continue;
    float d = lineDistance(origin, object->mPos);
    if ((d > maxrange+object->mRadius) || (d < minrange-object->mRadius)) continue;
    amount--;
    result->push_back(object);
  }
  return result;
}


list<cObject*>* cWorld::filterByName(cObject* ex, char* name, int maxamount, list<cObject*>* objects)
{
  list<cObject*>* result = new list<cObject*>;
  int amount = maxamount;
  if (objects == NULL) objects = &mObjects;
  loopObjects(*objects)
  {
    if (amount == 0) break;
    cObject *object = *it; it++;
    if (object->mFragged) continue;
    if (object == ex) continue;
    if (object->mName == NULL) continue;
    if (strcmp(object->mName->c_str(), name) != 0) continue;
    amount--;
    result->push_back(object);
  }
  return result;
}


list<cObject*>* cWorld::filterByCollideLine(cObject* ex, float* pointa, float* pointb, int maxamount, list<cObject*>* objects)
{
  list<cObject*>* result = new list<cObject*>;
  int amount = maxamount;
  if (objects == NULL) objects = &mObjects;
  loopObjects(*objects)
  {
    if (amount == 0) break;
    cObject *object = *it; it++;
    if (object->mFragged) continue;
    if (object == ex) continue;
    cModelCollision* collision = NULL;
    object->collideLine(pointa, pointb, collision);
    if (collision == NULL) continue;
    if (collision->intsec_entmod == NULL) continue;
    //cObject* obj = collision->intsec_entmod->mEntity;
    cObject* obj = object;
    amount--;
    result->push_back(obj);
  }
  return result;
}


list<cObject*>* cWorld::filterByCylinder(cObject* ex, float* pointa, float* pointb, float diameter, int maxamount, list<cObject*>* objects)
{
  float* x1 = pointa;
  float* x2 = pointb;
  list<cObject*>* result = new list<cObject*>;
  int amount = maxamount;
  if (objects == NULL) objects = &mObjects;
  loopObjects(*objects)
  {
    if (amount == 0) break;
    cObject *object = *it; it++;
    if (object->mFragged) continue;
    if (object == ex) continue;
    
    float* x0 = object->mPos;
    float u[] = { x0[0] - x1[0], x0[1] - x1[1], x0[2] - x1[2] };
    float v[] = { x2[0] - x1[0], x2[1] - x1[1], x2[2] - x1[2] };
    float a = u[0] * v[0] + u[1] * v[1] + u[2] * v[2];
    float b = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
    float delta = a/b;
    
    if (delta < -0.01) continue;
    if (delta > +1.01) continue;
    
    float p[] = {
        x1[0] + delta * (x2[0] - x1[0]),
        x1[1] + delta * (x2[1] - x1[1]),
        x1[2] + delta * (x2[2] - x1[2])
    };
    
    float dv[] = {
        p[0] - x0[0],
        p[1] - x0[1],
        p[2] - x0[2]
    };
    
    float d2 = dv[0]*dv[0] + dv[1]*dv[1] + dv[2]*dv[2];
    float is = diameter + object->mRadius;
    
    if (d2 > is*is) continue;
    
    amount--;
    result->push_back(object);
  }
  return result;
}


// Misc


const char* cWorld::getNATO(int letter)
{
    char* NATO[] =
    {
        "Alpha", "Bravo", "Charlie", "Delta", "Echo", "Foxtrot", "Golf", "Hotel", "India", "Juliett", "Kilo", "Lima", "Mike", "November", "Oscar", "Papa",
        "Quibec", "Romeo", "Sierra", "Tango", "Uniform", "Victor", "Whiskey", "X-Ray", "Yankee", "Zulu"
    };
}

